<?php

namespace InSegment\ApiCore\Traits;

use Illuminate\Support\Facades\App;

use InSegment\ApiCore\Facades\DTODefs;
use InSegment\ApiCore\Services\Transactor;
use InSegment\ApiCore\Interfaces\ApiKeyInterface;
use InSegment\ApiCore\Api\DataTransferObject;

use Symfony\Component\HttpFoundation\Response;

trait ProvidesApiEntryPoints
{
    /**
     * Authorization through ApiKey in JSON
     * 
     * If the Api Key is failed to set, the response status will be 401 Unauthorized
     * 
     * @param ApiKeyInterface $apiKey
     * @return array
     */
    public function authorizeApiKey(ApiKeyInterface $apiKey)
    {
        app('session')->put(['api_key' => $apiKey->getPublicKey()]);
        
        return $this->apiResponse();
    }

    /**
     * Get or create Transaction
     * 
     * The format of input:
     * {
     *   "root": "App\\Campaign",
     *   "count": {
     *     "campaigns": ...,
     *     "campaign_templates": ...,
     *     "campaign_template_mappings": ...,
     *     "contacts_template_mapping": ...,
     *     "companies":...,
     *     "company_ranges": ...,
     *     "company_industries": ...,
     *     "lists": ...,
     *     "contacts": ...,
     *     "contact_template_mappings": ...,
     *     "contact_titles": ...,
     *     "contact_phones": ...,
     *     "proof_links": ...
     *   }
     * }
     * 
     * @return array
     */
    public function acquireTransaction(ApiKeyInterface $apiKey)
    {
        $data = app('request')->all();
        $rootClass = data_get($data, 'root', '');
        $rootDTOs = DTODefs::getGlobalRootDTOs();
        
        if (!$rootClass || !in_array($rootClass, $rootDTOs)) {
            $rootClass = $rootDTOs[0];
        }
        
        Transactor::prepare($rootClass, data_get($data, 'count', []));
        
        return $this->apiResponse();
    }

    /**
     * Close Transaction
     * 
     * @return array
     */
    public function closeTransaction(ApiKeyInterface $apiKey)
    {
        $data = app('request')->all();
        $clear = data_get($data, 'clear', true);
        $result = Transactor::watchingTables(function (Transactor $transactor) use ($clear) {
            $mergeOutput = DTODefs::onTransactionMerge($transactor);
            
            return $transactor->commit(function (Transactor $transactor) use ($mergeOutput) {
                return DTODefs::postProcessTransaction($transactor, $mergeOutput);
            }, $clear);
        });

        return $this->apiResponse($result);
    }
    
    /**
     * Undo Transaction
     * 
     * @param ApiKeyInterface $apiKey
     * @return array
     */
    public function showTransactionStatus(ApiKeyInterface $apiKey)
    {
        $columns = data_get(app('request')->all(), 'columns');
        return $this->apiResponse(Transactor::status($columns), 'status');
    }

    /**
     * Undo Transaction
     * 
     * @param ApiKeyInterface $apiKey
     * @return array
     */
    public function undoTransaction(ApiKeyInterface $apiKey)
    {
        Transactor::watchingTables(function (Transactor $transactor) {
            return $transactor->undo();
        });

        return $this->apiResponse();
    }

    /**
     * Set/Add Object
     * 
     * @param string $type
     * @param ApiKeyInterface $apiKey
     * @return array
     */
    public function set($type, ApiKeyInterface $apiKey)
    {
        return $this->import($type, false);
    }
    
    /**
     * Update Object
     * 
     * @param string $type
     * @param ApiKeyInterface $apiKey
     * @return array
     */
    public function update($type, ApiKeyInterface $apiKey)
    {
        return $this->import($type, true);
    }
    
    /**
     * Set or update Object
     * 
     * @param string $type
     * @param bool $update
     * @return array
     */
    private function import($type, bool $update)
    {
        $data = app('request')->json()->all();
        
        $result = Transactor::watchingTables(function (Transactor $transactor) use (&$type, &$data, &$update) {
            $ret = (new DataTransferObject($type, $data, str_singular($type)))->processIncoming($update);
            $transactor->getBufferedWriter()->flushAll();
            return $ret;
        });
        
        if ($result === DataTransferObject::DTO_IMPORT_REQUIRE) {
            return $this->apiResponse(str_singular($type), 'require');
        }
        
        return $this->apiResponse();
    }
    
    /**
     * Export an object
     * 
     * @param string $type
     * @param ApiKeyInterface $apiKey
     * @return array
     */
    public function export($type, ApiKeyInterface $apiKey)
    {
        $data = app('request')->json()->all();
        
        $result = Transactor::watchingTables(function (Transactor $transactor) use (&$type, &$data) {
            return (new DataTransferObject($type, $data))->processOutgoing();
        });
        
        return $this->apiResponse($result, 'data');
    }
    
    /**
     * Return response
     *
     * @param array|string $data response data
     * @param string $root
     *
     * @return array
     */
    public function apiResponse($data = null, string $root = 'data')
    {
        $responseData = ['status' => Response::HTTP_OK];
        
        if (isset($data)) {
            array_set($responseData, $root, $data);
        }
        
        return response()->json($responseData, Response::HTTP_OK);
    }
}
